/*
	Copyright (C) 2004 - 2007 rt2x00 SourceForge Project
	<http://rt2x00.serialmonkey.com>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the
	Free Software Foundation, Inc.,
	59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
	Module: rt2x00lib
	Abstract: rt2x00 generic pci device routines.
	Supported chipsets: RT2460, RT2560, rt2561, rt2561s & rt2661.
 */

#include <linux/pci.h>

#include "rt2x00.h"
#include "rt2x00lib.h"
#include "rt2x00pci.h"

/*
 * Beacon handlers.
 */
void rt2x00pci_beacon_update(struct ieee80211_hw *hw, const int queue,
	struct sk_buff *skb, struct ieee80211_tx_control *control)
{
	struct rt2x00_dev *rt2x00dev = hw->priv;
	struct data_ring *ring = &rt2x00dev->ring[queue];
	struct data_entry *entry = rt2x00_get_data_entry(ring);

	/*
	 * Just in case the ieee80211 doesn't set this,
	 * but we need this queue set for the descriptor
	 * initialization.
	 */
	control->queue = IEEE80211_TX_QUEUE_BEACON;

	/*
	 * Update the beacon entry.
	 */
	memcpy(entry->data_addr, skb->data, skb->len);
	rt2x00dev->lib_ops->write_tx_desc(rt2x00dev, entry->priv,
		(struct ieee80211_hdr *)skb->data, skb->len, control);

}
EXPORT_SYMBOL_GPL(rt2x00pci_beacon_update);

void rt2x00pci_beacondone(struct rt2x00_dev *rt2x00dev, const int queue)
{
	struct data_ring *ring = &rt2x00dev->ring[queue];
	struct data_entry *entry = rt2x00_get_data_entry(ring);
	struct sk_buff *skb;

	skb = ieee80211_beacon_get(rt2x00dev->hw,
		rt2x00dev->interface.id, &entry->tx_status.control);
	if (!skb)
		return;

	rt2x00dev->hw_ops->beacon_update(rt2x00dev->hw, skb,
		&entry->tx_status.control);

	dev_kfree_skb(skb);
}
EXPORT_SYMBOL_GPL(rt2x00pci_beacondone);

/*
 * Device initialization handlers.
 */
#define priv_offset(__ring, __i)				\
({								\
	ring->data_addr + (i * ring->desc_size);		\
})

#define data_addr_offset(__ring, __i)				\
({								\
	(__ring)->data_addr					\
		+ ((__ring)->stats.limit * (__ring)->desc_size)	\
		+ ((__i) * (__ring)->data_size);		\
})

#define data_dma_offset(__ring, __i)				\
({								\
	(__ring)->data_dma					\
		+ ((__ring)->stats.limit * (__ring)->desc_size)	\
		+ ((__i) * (__ring)->data_size);		\
})

static int rt2x00pci_alloc_dma_ring(struct rt2x00_dev *rt2x00dev,
	const int queue, const u16 max_entries,
	const u16 data_size, const u16 desc_size)
{
	struct data_ring *ring = &rt2x00dev->ring[queue];
	unsigned int i;

	ring->stats.limit = max_entries;
	ring->data_size = data_size;
	ring->desc_size = desc_size;

	/*
	 * Allocate all ring entries.
	 */
	ring->entry = kzalloc(ring->stats.limit * sizeof(struct data_entry),
		GFP_KERNEL);
	if (!ring->entry)
		return -ENOMEM;

	/*
	 * Allocate DMA memory for descriptor and buffer.
	 */
	ring->data_addr = pci_alloc_consistent(rt2x00dev_pci(rt2x00dev),
		rt2x00_get_ring_size(ring), &ring->data_dma);
	if (!ring->data_addr) {
		kfree(ring->entry);
		return -ENOMEM;
	}

	/*
	 * Initialize all ring entries to contain valid
	 * addresses.
	 */
	for (i = 0; i < ring->stats.limit; i++) {
		ring->entry[i].flags = 0;
		ring->entry[i].ring = ring;
		ring->entry[i].skb = NULL;
		ring->entry[i].priv = priv_offset(ring, i);
		ring->entry[i].data_addr = data_addr_offset(ring, i);
		ring->entry[i].data_dma = data_dma_offset(ring, i);
	}

	return 0;
}

static void rt2x00pci_free_ring(struct rt2x00_dev *rt2x00dev, const int queue)
{
	struct data_ring *ring = &rt2x00dev->ring[queue];
	if (ring->data_addr)
		pci_free_consistent(rt2x00dev_pci(rt2x00dev),
			rt2x00_get_ring_size(ring),
			ring->data_addr, ring->data_dma);
	ring->data_addr = NULL;

	kfree(ring->entry);
	ring->entry = NULL;
}

static int rt2x00pci_allocate_dma_rings(struct rt2x00_dev *rt2x00dev)
{
	unsigned int i;
	int rings;

	/*
	 * The TX rings reported to mac80211,
	 * plus the additional ATIM ring (if supported).
	 */
	rings = rt2x00dev->hw->queues +
		GET_FLAG(rt2x00dev, DEVICE_SUPPORT_ATIM);

	/*
	 * First allocate the TX rings (minus BEACON).
	 */
	for (i = 0; i < rings; i++) {
		if (rt2x00pci_alloc_dma_ring(rt2x00dev, i,
			TX_ENTRIES, DATA_FRAME_SIZE, rt2x00dev->txd_size))
			return -ENOMEM;
	}

	/*
	 * Allocate the BEACON ring.
	 */
	if (rt2x00pci_alloc_dma_ring(rt2x00dev, rings,
		BEACON_ENTRIES, MGMT_FRAME_SIZE, rt2x00dev->txd_size))
		return -ENOMEM;

	/*
	 * Allocate the RX ring.
	 */
	if (rt2x00pci_alloc_dma_ring(rt2x00dev, rings + 1,
		RX_ENTRIES, DATA_FRAME_SIZE, rt2x00dev->rxd_size))
		return -ENOMEM;

	return 0;
}

static void rt2x00pci_free_rings(struct rt2x00_dev *rt2x00dev)
{
	unsigned int i;
	int rings;

	/*
	 * The BEACON and RX ring, plus TX rings reported to mac80211,
	 * and the additional ATIM ring (if supported).
	 */
	rings = 2 + rt2x00dev->hw->queues +
		GET_FLAG(rt2x00dev, DEVICE_SUPPORT_ATIM);

	for (i = 0; i < rings; i++)
		rt2x00pci_free_ring(rt2x00dev, i);
}

int rt2x00pci_initialize(struct rt2x00_dev *rt2x00dev)
{
	unsigned int irq;
	int status;

	if (GET_FLAG(rt2x00dev, DEVICE_INITIALIZED))
		return 0;

	/*
	 * We must wait on the firmware before
	 * we can safely continue.
	 */
	if (rt2x00lib_load_firmware_wait(rt2x00dev))
		return -ENODEV;

	/*
	 * Allocate all data rings.
	 */
	status = rt2x00pci_allocate_dma_rings(rt2x00dev);
	if (status) {
		ERROR("DMA allocation failed.\n");
		goto exit_fail;
	}

	/*
	 * Reset the channel_change_time value
	 * to make sure it will be correctly initialized
	 * after the radio has been enabled.
	 */
	rt2x00dev->hw->channel_change_time = 0;

	/*
	 * Register interrupt handler.
	 */
	irq = rt2x00dev_pci(rt2x00dev)->irq;
	status = request_irq(irq, rt2x00dev->lib_ops->irq_handler,
		IRQF_SHARED, DRV_NAME, rt2x00dev);
	if (status) {
		ERROR("IRQ %d allocation failed (error %d).\n", irq, status);
		goto exit_fail;
	}

	SET_FLAG(rt2x00dev, DEVICE_INITIALIZED);

	return 0;

exit_fail:
	rt2x00pci_free_rings(rt2x00dev);

	return status;
}
EXPORT_SYMBOL_GPL(rt2x00pci_initialize);

void rt2x00pci_uninitialize(struct rt2x00_dev *rt2x00dev)
{
	if (!GET_FLAG(rt2x00dev, DEVICE_INITIALIZED))
		return;

	/*
	 * Cancel scanning.
	 */
	if (rt2x00dev->scan)
		rt2x00_signal_scan(rt2x00dev->scan, SCANNING_CANCELLED);

	/*
	 * Flush out all pending work.
	 */
	flush_workqueue(rt2x00dev->workqueue);

	/*
	 * Free DMA rings.
	 */
	rt2x00pci_free_rings(rt2x00dev);

	/*
	 * Free irq line.
	 */
	free_irq(rt2x00dev_pci(rt2x00dev)->irq, rt2x00dev);

	CLEAR_FLAG(rt2x00dev, DEVICE_INITIALIZED);
}
EXPORT_SYMBOL_GPL(rt2x00pci_uninitialize);
